@@ -27,10 +27,15 @@
 #include <initguid.h>
 #include "timiditydrv.h"
 
-#include "mmddk.h"
-#include <mmsystem.h>
+#include "mmddk.h"  //Fom NTDDK
+/* Use XP SDK's mmsystem.h. mingw's lacks some definitions */
+#include "mmsystem.h"
+
+#ifdef __MINGW32__
+#define __IID_DEFINED__ 1
+#endif
 #include "timiditydrv_i.c"
-#include <mmreg.h>
+#include "mmreg.h"	//Fom NTDDK
 extern "C" {
 #include "config.h"
 #include "sysdep.h"
@@ -40,12 +45,14 @@ extern "C" {
 
 LONG driverCount;
 
-static int OpenCount = 0;
+static volatile int OpenCount = 0;
+static volatile int modm_closed = 1;
 
 static CRITICAL_SECTION mim_section;
-static int stop_thread = 0;
+static volatile int stop_thread = 0;
+static volatile int stop_rtthread = 0;
 static HANDLE hCalcThread = NULL;
-
+static HANDLE hRtsynThread =NULL;
 
 
 extern "C"
@@ -249,15 +256,14 @@ int timsyn_play_some_data(void){
 }
 
 DWORD WINAPI threadfunc(LPVOID lpV){
-	int stopt = 0;
-	while(stopt == 0){
+	while(stop_thread == 0){
 		Sleep(1);
 		EnterCriticalSection(&mim_section);
 		timsyn_play_some_data();
 		rtsyn_play_calculate();
-		stopt=stop_thread;
 		LeaveCriticalSection(&mim_section);
 	}
+	stop_thread=0;
 	ExitThread(TRUE);
 }
 
@@ -266,13 +272,9 @@ DWORD WINAPI threadfunc2(LPVOID lpV){
 	char *argv[2];
 	HANDLE hThread = NULL;
 	DWORD ThreadID;
-	int OCount;
-	int stopt = 0;
-	
-	EnterCriticalSection(&mim_section);
-	OCount=OpenCount;
-	LeaveCriticalSection(&mim_section);
-	while(OCount <= 0) {
+	int opend=0;
+	;
+	while(opend == 0) {
 		Sleep(100);
 		argc = 2;
 		argv[0] = strdup("timidity");
@@ -281,47 +283,40 @@ DWORD WINAPI threadfunc2(LPVOID lpV){
 			rtsyn_init();
 			hThread=CreateThread(NULL,0,threadfunc,0,0,&ThreadID);
 			SetPriorityClass(hThread, REALTIME_PRIORITY_CLASS);
-			OCount = 1;
+			opend = 1;
 		}
 		for(i = 0 ; i < 2 ; i++) free(argv[i]);
 	}
-	EnterCriticalSection(&mim_section);
 	hCalcThread = hThread;
-	OpenCount=OCount;
-	LeaveCriticalSection(&mim_section);
 	
-	while(stopt == 0){
-		Sleep(100);
-		EnterCriticalSection(&mim_section);
-		stopt=stop_thread;
-		LeaveCriticalSection(&mim_section);
+	while(stop_rtthread == 0){
+		Sleep(10);
 	}
+
 	rtsyn_stop_playing();
 	rtsyn_close();
 	timiwp_main_close();
+	stop_rtthread=0;
 	ExitThread(TRUE);
 }
 
 STDAPI_(LONG) modMessage(UINT uDeviceID, UINT uMsg, DWORD dwUser, DWORD dwParam1, DWORD dwParam2){
-	HANDLE hThread = NULL;
 	DWORD ThreadID;
 	DWORD tstate;
 	int OCount;
+	DWORD Exit;
 	
 	MIDIHDR *IIMidiHdr;	
 	UINT evbpoint;
+
+
 	
 	switch (uMsg) {
 	case MODM_OPEN:
-		EnterCriticalSection(&mim_section);
-		OCount=OpenCount;
-		LeaveCriticalSection(&mim_section);
-		if ( OCount <= 0 ){
-			hThread=CreateThread(NULL,0,threadfunc2,0,0,&ThreadID);
-		}else{
-			EnterCriticalSection(&mim_section);
-			OpenCount++;
-			LeaveCriticalSection(&mim_section);
+		OpenCount++;
+		if ( OpenCount == 1 && modm_closed  == 1 ){
+			hRtsynThread=CreateThread(NULL,0,threadfunc2,0,0,&ThreadID);
+			modm_closed = 0;
 		}
 		return MMSYSERR_NOERROR;
 		break;
@@ -355,24 +350,33 @@ STDAPI_(LONG) modMessage(UINT uDeviceID, UINT uMsg, DWORD dwUser, DWORD dwParam1
 		return 0x1;
 		break;
 	case MODM_CLOSE:
-		EnterCriticalSection(&mim_section);
-		OCount=--OpenCount;
-		LeaveCriticalSection(&mim_section);
-		if( OCount == 0){
+		if ( stop_rtthread != 0 || stop_thread != 0 ) return MIDIERR_STILLPLAYING;
+		--OpenCount;
+		if( OpenCount == 1){
 			int maxloop=1000;
-			EnterCriticalSection(&mim_section);
-			stop_thread = ~0;
-			LeaveCriticalSection(&mim_section);
-			do{
-				Sleep(1);
-				EnterCriticalSection(&mim_section);
-				GetExitCodeThread(hCalcThread , &tstate);
-				LeaveCriticalSection(&mim_section);
-			}while(tstate == STILL_ACTIVE && maxloop-- > 0);
-			EnterCriticalSection(&mim_section);
+			
+			stop_thread = 1;
+			while( stop_thread != 0 && maxloop-- > 0) Sleep(1);
+			if(stop_thread == 0) {
+				stop_rtthread = 1;
+				while( stop_rtthread != 0 && maxloop-- > 0) Sleep(1);
+			}
+			if(stop_rtthread != 0) TerminateThread(hRtsynThread, GetExitCodeThread(hRtsynThread,&Exit));
+			if(stop_thread != 0) TerminateThread(hCalcThread, GetExitCodeThread(hCalcThread,&Exit));
+			if(maxloop == 0){
+				DeleteCriticalSection(&mim_section);
+				InitializeCriticalSection(&mim_section);
+			}
+			stop_rtthread = 0;
 			stop_thread = 0;
-			LeaveCriticalSection(&mim_section);
+			CloseHandle(hRtsynThread);
 			CloseHandle(hCalcThread);
+			modm_closed=1;
+		}else{ 
+			if(OpenCount < 0){
+				OpenCount = 0;
+				return MMSYSERR_NOTENABLED;
+			}
 		}
 		return MMSYSERR_NOERROR;
 		break;
